//# tCorrelatorSubbandProcProcessSb.cc 
//#
//# Copyright (C) 2013  ASTRON (Netherlands Institute for Radio Astronomy)
//# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
//#
//# This file is part of the LOFAR software suite.
//# The LOFAR software suite is free software: you can redistribute it and/or
//# modify it under the terms of the GNU General Public License as published
//# by the Free Software Foundation, either version 3 of the License, or
//# (at your option) any later version.
//#
//# The LOFAR software suite is distributed in the hope that it will be useful,
//# but WITHOUT ANY WARRANTY; without even the implied warranty of
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//# GNU General Public License for more details.
//#
//# You should have received a copy of the GNU General Public License along
//# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
//#
//# $Id$

#include <lofar_config.h>

#include <Common/LofarLogger.h>
#include <Common/LofarTypes.h>
#include <CoInterface/fpequals.h>
#include <CoInterface/Parset.h>
#include <GPUProc/gpu_utils.h>
#include <GPUProc/SubbandProcs/KernelFactories.h>
#include <GPUProc/SubbandProcs/SubbandProc.h>

#include <UnitTest++.h>

using namespace std;
using namespace LOFAR;
using namespace LOFAR::Cobalt;

struct SubbandProcWrapper {
  // Create very simple kernel programs, with predictable output. Skip as much
  // as possible. Nr of channels/sb from the parset is 1, so the PPF will not
  // even run.  Parset also has turned of delay compensation and bandpass
  // correction (but that kernel will run to convert int to float and to
  // transform the data order).

  gpu::Device device;
  gpu::Context ctx;

  Parset ps;
  KernelFactories factories;
  SubbandProc cwq;
  SubbandProcInputData in;
  SubbandProcOutputData out;

  fcomplex inputValue;

  SubbandProcWrapper(Parset &ps):
    device(0),
    ctx(device),
    ps(ps),
    factories(ps, 1),
    cwq(ps, ctx, factories),
    in(ps.settings.SAPs.size(),
       ps.settings.antennaFields.size(),
       ps.settings.nrPolarisations,
       ps.settings.beamFormer.maxNrTABsPerSAP(),
       ps.settings.blockSize,
       ps.nrBytesPerComplexSample(),
       ctx),
    out(ps, ctx),

    inputValue(1,1)
  {
    // Input info
    const size_t nrBeams = ps.settings.SAPs.size();
    const size_t nrStations = ps.settings.antennaFields.size();
    const size_t nrPolarisations = ps.settings.nrPolarisations;
    const size_t maxNrTABsPerSAP = ps.settings.beamFormer.maxNrTABsPerSAP();
    const size_t nrSamplesPerIntegration = ps.settings.correlator.nrSamplesPerIntegration();
    const size_t nrSamplesPerSubband = ps.settings.blockSize;
    const size_t nrBitsPerSample = ps.settings.nrBitsPerSample;
    const size_t nrBytesPerComplexSample = ps.nrBytesPerComplexSample();

    // We only support 8-bit or 16-bit input samples
    ASSERT(nrBitsPerSample == 8 || nrBitsPerSample == 16);

    // Output info
    const size_t nrBaselines = ps.nrBaselines();
    const size_t nrBlocksPerIntegration = 
      ps.settings.correlator.nrBlocksPerIntegration;
    const size_t nrChannelsPerSubband = ps.settings.correlator.nrChannels;
    const size_t integrationSteps = ps.settings.correlator.nrSamplesPerIntegration();

    LOG_INFO_STR(
      "\nInput info:" <<
      "\n  nrBeams = " << nrBeams <<
      "\n  nrStations = " << nrStations <<
      "\n  nrPolarisations = " << nrPolarisations <<
      "\n  maxNrTABsPerSAP = " << maxNrTABsPerSAP <<
      "\n  nrSamplesPerIntegration = " << nrSamplesPerIntegration <<
      "\n  nrSamplesPerSubband = " << nrSamplesPerSubband <<
      "\n  nrBitsPerSample = " << nrBitsPerSample <<
      "\n  nrBytesPerComplexSample = " << nrBytesPerComplexSample << 
      "\n  ----------------------------" <<
      "\n  Total bytes = " << in.inputSamples.size());

    LOG_INFO_STR(
      "\nOutput info:" <<
      "\n  nrBaselines = " << nrBaselines <<
      "\n  nrBlockPerIntegration = " << nrBlocksPerIntegration << 
      "\n  nrChannelsPerSubband = " << nrChannelsPerSubband <<
      "\n  integrationSteps = " << integrationSteps <<
      "\n  ----------------------------" <<
      "\n  Total bytes = " << out.correlatedData.subblocks[0]->visibilities.size());

    // Initialize synthetic input to all (1, 1).
    for (size_t st = 0; st < nrStations; st++)
      for (size_t i = 0; i < nrSamplesPerSubband; i++)
        for (size_t pol = 0; pol < nrPolarisations; pol++)
          setInputValue(st, i, pol, inputValue);

    // Initialize subbands partitioning administration (struct BlockID). We only
    // do the 1st block of whatever.
    in.blockID.block = 0;
    in.blockID.globalSubbandIdx = 0;
    in.blockID.localSubbandIdx = 0;
    in.blockID.subbandProcSubbandIdx = 0;

    // Initialize delays. We skip delay compensation, but init anyway,
    // so we won't copy uninitialized data to the device.
    for (size_t i = 0; i < in.delaysAtBegin.size(); i++)
      in.delaysAtBegin.get<float>()[i] = 0.0f;
    for (size_t i = 0; i < in.delaysAfterEnd.size(); i++)
      in.delaysAfterEnd.get<float>()[i] = 0.0f;
    for (size_t i = 0; i < in.phase0s.size(); i++)
      in.phase0s.get<float>()[i] = 0.0f;
  }

  void setInputValue(unsigned station, unsigned t, unsigned pol, fcomplex value) {
    const size_t nrBitsPerSample = ps.settings.nrBitsPerSample;

    switch(nrBitsPerSample) {
    case 8:
      reinterpret_cast<i8complex&>(in.inputSamples[station][t][pol][0]) = 
        i8complex(value);
      break;
    case 16:
      reinterpret_cast<i16complex&>(in.inputSamples[station][t][pol][0]) = 
        i16complex(value);
      break;
    }
  }

  void process() {
    const ssize_t nrBlocksPerIntegration = 
      ps.settings.correlator.nrBlocksPerIntegration;

    ssize_t block(0);

    LOG_INFO("Processing ...");
    for (block = -1; block < nrBlocksPerIntegration; block++) {
      LOG_DEBUG_STR("Processing block #" << block);

      in.blockID.block = block;
      cwq.processSubband(in, out);
      if (block >= 0)
        cwq.postprocessSubband(out);
    }
    ASSERT(block == nrBlocksPerIntegration);
    ASSERT(out.emit_correlatedData);
  }

  fcomplex outputValue() const {
    // The output is the correlation-product of two inputs (with identical
    // `inputValue`).

    const size_t scaleFactor = ps.settings.nrBitsPerSample == 16 ? 1 : 16;

    return norm(inputValue) * scaleFactor * scaleFactor;
  }

  void verifyOutput(const fcomplex expectedValue) const {
    // we don't process the FFT in our reference calculations yet
    ASSERT(ps.settings.correlator.nrChannels == 1);

    LOG_INFO("Verifying output ...");
    for (size_t b = 0; b < ps.nrBaselines(); b++)
      for (size_t c = 0; c < ps.settings.correlator.nrChannels; c++)
        for (size_t pol0 = 0; pol0 < ps.settings.nrPolarisations; pol0++)
          for (size_t pol1 = 0; pol1 < ps.settings.nrPolarisations; pol1++)
            ASSERTSTR(fpEquals(out.correlatedData.subblocks[0]->visibilities[b][c][pol0][pol1], expectedValue),
                      "out[" << b << "][" << c << "][" << pol0 << 
                      "][" << pol1 << "] = " << out.correlatedData.subblocks[0]->visibilities[b][c][pol0][pol1] << 
                      "; expectedValue = " << expectedValue);
  }
};

// Test the output on clean data -- should produce normal output
TEST(output_noflags_uniform) {
  Parset  ps("tCorrelatorSubbandProcProcessSb.parset");
  SubbandProcWrapper wrapper(ps);

  wrapper.process();
  wrapper.verifyOutput(wrapper.outputValue());
}

// Test the output on clean data with a peak -- should produce elevated output
TEST(output_noflags_peak) {
  Parset  ps("tCorrelatorSubbandProcProcessSb.parset");
  SubbandProcWrapper wrapper(ps);

  // Replace one value with an extreme value to know whether it's actually skipped
  for(size_t st = 0; st < ps.settings.antennaFields.size(); st++) {
    wrapper.setInputValue(st, 13, 0, fcomplex(100,0));
    wrapper.setInputValue(st, 13, 1, fcomplex(100,0));
  }

  // Each unmodified input value of 1+1i produces a correlation of (1+1i)(1-1i) = 2
  // Each modified input value of 100 produces a correlation of 100 * 100 = 1e4
  // All correlations are averaged and scaled up.

  wrapper.process();
  wrapper.verifyOutput(3011.5); // = ((blockSize-1) * 2 + 1e4) / blockSize * scaleFactor * scaleFactor
}

// Test the output on clean data with a flaged peak -- should produce normal output
TEST(output_flags) {
  Parset  ps("tCorrelatorSubbandProcProcessSb.parset");
  SubbandProcWrapper wrapper(ps);

  // Flag one sample
  wrapper.in.inputFlags[0].include(13);
  // Replace it with an extreme value to know whether it's actually skipped
  wrapper.setInputValue(0, 13, 0, fcomplex(100,100));
  wrapper.setInputValue(0, 13, 1, fcomplex(100,100));

  // process
  wrapper.process();
  wrapper.verifyOutput(wrapper.outputValue());
}

// Test the final weights if there is flagged data
TEST(weights_flags_1ch) {
  // Override nr channels to 1
  Parset  ps("tCorrelatorSubbandProcProcessSb.parset");
  ps.replace("Cobalt.Correlator.nrChannelsPerSubband", "1");
  ps.updateSettings();
  SubbandProcWrapper wrapper(ps);

  // Flag one sample
  wrapper.in.inputFlags[0].include(13);

  // process
  wrapper.process();

  // We reuse the flagged sample in all of the blocks, so
  // we actually lose "nrBlocks" samples.
  const unsigned nrBlocks = ps.settings.correlator.nrBlocksPerIntegration;
  const unsigned nrValidSamples = wrapper.ps.settings.correlator.nrSamplesPerIntegration();
  const unsigned nrValidSamplesFlagged =
    nrValidSamples
    // we lose 1 sample per block for each flagged sample
    - nrBlocks * 1;

  LOG_INFO("Verifying output weights ...");
  for (size_t b = 0; b < wrapper.ps.nrBaselines(); b++) {
    // baseline 0 and 1 contain station 0 with a flagged sample.
    unsigned expected = b < 2 ? nrValidSamplesFlagged : nrValidSamples;
    ASSERTSTR(wrapper.out.correlatedData.subblocks[0]->getNrValidSamples(b, 0) == expected,
              "nrValidSamples[" << b << "][0] = "
              << wrapper.out.correlatedData.subblocks[0]->getNrValidSamples(b, 0)
              << "; expected " << expected);
  }
}

// Test the final weights after FFT if there is flagged data
TEST(weights_flags_64ch) {
  // Override nr channels to 64
  Parset  ps("tCorrelatorSubbandProcProcessSb.parset");
  ps.replace("Cobalt.Correlator.nrChannelsPerSubband", "64");
  ps.updateSettings();
  SubbandProcWrapper wrapper(ps);

  // Flag one sample
  wrapper.in.inputFlags[0].include(13);

  // process
  wrapper.process();

  // We reuse the flagged sample in all of the blocks, so
  // we actually lose "nrBlocks" samples.
  const unsigned nrBlocks = ps.settings.correlator.nrBlocksPerIntegration;
  const unsigned nrValidSamples = wrapper.ps.settings.correlator.nrSamplesPerIntegration();
  const unsigned nrValidSamplesFlagged =
    nrValidSamples
    // we lose "NR_TAPS" samples per block for each flagged sample
    - nrBlocks * NR_TAPS;

  LOG_INFO("Verifying output weights ...");
  for (size_t b = 0; b < wrapper.ps.nrBaselines(); b++)
    for (size_t c = 1; c < wrapper.ps.settings.correlator.nrChannels; c++) {
      // baseline 0 and 1 contain station 0 with a flagged sample.
      unsigned expected = b < 2 ? nrValidSamplesFlagged : nrValidSamples;
      ASSERTSTR(wrapper.out.correlatedData.subblocks[0]->getNrValidSamples(b, c) == expected,
                "nrValidSamples[" << b << "][" << c << "] = "
                << wrapper.out.correlatedData.subblocks[0]->getNrValidSamples(b, c)
                << "; expected " << expected);
    }
}

int main() {
  INIT_LOGGER("tCorrelatorSubbandProcProcessSb");

  try {
    gpu::Platform pf;
    return UnitTest::RunAllTests() > 0;
  } catch (gpu::CUDAException& e) {
    cerr << e.what() << endl;
    return 3;
  }
}

